/*	
	Papuc Costin Giorgian
	Grupa 333CB
	Proiect PM
 */

#include <avr/io.h>
#include <util/delay.h>

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>

#define SIZE 4

uint8_t findTarget(uint8_t array[SIZE],uint8_t x,uint8_t stop) {
	uint8_t t;
	// if the position is already on the first, don't evaluate
	if (x == 0) {
		return x;
	}
	for(t = x - 1; t >= 0; t--) {
		if (array[t] != 0) {
			if (array[t] != array[x]) {
				// merge is not possible, take next position
				return t + 1;
			}
			return t;
		} else {
			// we should not slide further, return this one
			if (t == stop) {
				return t;
			}
		}
	}
	// we did not find a
	return x;
}

bool slideArray(uint8_t array[SIZE]) {
	bool success = false;
	uint8_t x, t, stop=0;

	for (x = 0; x < SIZE; x++) {
		if (array[x] != 0) {
			t = findTarget(array,x,stop);
			// if target is not original position, then move or merge
			if (t != x) {
				// if target is zero, this is a move
				if (array[t] == 0) {
					array[t] = array[x];
				} else if (array[t] == array[x]) {
					// merge (increase power of two)
					array[t]++;
					// set stop to avoid double merge
					stop = t+1;
				}
				array[x] = 0;
				success = true;
			}
		}
	}
	return success;
}

void rotateBoard(uint8_t board[SIZE][SIZE]) {
	uint8_t i, j, n = SIZE;
	uint8_t tmp;
	for (i = 0; i < n / 2; i++) {
		for (j = i; j < n-i-1; j++) {
			tmp = board[i][j];
			board[i][j] = board[j][n-i-1];
			board[j][n-i-1] = board[n-i-1][n-j-1];
			board[n-i-1][n-j-1] = board[n-j-1][i];
			board[n-j-1][i] = tmp;
		}
	}
}

bool moveUp(uint8_t board[SIZE][SIZE]) {
	bool success = false;
	uint8_t x;
	for (x=0;x<SIZE;x++) {
		success |= slideArray(board[x]);
	}
	return success;
}

bool moveLeft(uint8_t board[SIZE][SIZE]) {
	bool success;
	rotateBoard(board);
	success = moveUp(board);
	rotateBoard(board);
	rotateBoard(board);
	rotateBoard(board);
	return success;
}

bool moveDown(uint8_t board[SIZE][SIZE]) {
	bool success;
	rotateBoard(board);
	rotateBoard(board);
	success = moveUp(board);
	rotateBoard(board);
	rotateBoard(board);
	return success;
}

bool moveRight(uint8_t board[SIZE][SIZE]) {
	bool success;
	rotateBoard(board);
	rotateBoard(board);
	rotateBoard(board);
	success = moveUp(board);
	rotateBoard(board);
	return success;
}

bool findPairDown(uint8_t board[SIZE][SIZE]) {
	uint8_t x, y;
	for (x = 0;x < SIZE; x++) {
		for (y = 0; y < SIZE - 1; y++) {
			if (board[x][y] == board[x][y+1])
				return true;
		}
	}
	for (x = 0;x < SIZE - 1; x++) {
		for (y = 0; y < SIZE; y++) {
			if (board[x][y] == board[x+1][y])
				return true;
		}
	}
	return false;
}

bool search256(uint8_t board[SIZE][SIZE]) {
	uint8_t x, y;

	for (x = 0; x < SIZE; x++) {
		for (y = 0; y < SIZE; y++) {
			if (board[x][y] == 7) {
				return true;
			}
		}
	}
	return false;
}

uint8_t countEmpty(uint8_t board[SIZE][SIZE]) {
	uint8_t x, y;
	uint8_t count = 0;
	for (x = 0; x < SIZE; x++) {
		for (y = 0; y < SIZE; y++) {
			if (board[x][y] == 0) {
				count++;
			}
		}
	}
	return count;
}

uint8_t gameEnded(uint8_t board[SIZE][SIZE]) {
	uint8_t ended = 1;
	if (search256(board))
		return 2;
	if (countEmpty(board) > 0)
		return 0;
	if (findPairDown(board))
		return 0;
	rotateBoard(board);
	if (findPairDown(board))
		ended = 0;
	rotateBoard(board);
	rotateBoard(board);
	rotateBoard(board);
	return ended;
}

void addRandom(uint8_t board[SIZE][SIZE]) {
	static bool initialized = false;
	uint8_t x, y;
	uint8_t r, len = 0;
	uint8_t n, list[SIZE*SIZE][2];

	if (!initialized) {
		srand(1234);
		initialized = true;
	}

	for (x = 0; x < SIZE; x++) {
		for (y = 0; y < SIZE; y++) {
			if (board[x][y] == 0) {
				list[len][0] = x;
				list[len][1] = y;
				len++;
			}
		}
	}

	if (len>0) {
		r = rand() % len;
		x = list[r][0];
		y = list[r][1];
		//n = (rand() % 10) / 9 + 1;
		board[x][y] = 1;
	}
}

void initBoard(uint8_t board[SIZE][SIZE]) {
	uint8_t x, y;
	for (x = 0; x < SIZE; x++) {
		for (y=0; y < SIZE; y++) {
			board[x][y] = 0;
		}
	}
	addRandom(board);
	addRandom(board);
}

int main() {
	uint8_t board[SIZE][SIZE];
	uint8_t i, j;
	uint8_t f;
	uint8_t max = 0;
	bool success = false;
	bool prim = false;

	initBoard(board);
	DDRB = 255;
	DDRD = 255;

	DDRA |= ~((1 << PA5) | (1 << PB6) | (1 << PA7));
	DDRA &= ~((1 << PA1) | (1 << PB2) | (1 << PA3) | (1 << PB4));
    PORTA |= ((1 << PA1) | (1 << PB2) | (1 << PA3) | (1 << PB4));
	uint8_t aux = PORTA;
	while (true) {
		
		PORTB = 0;
		PORTD = 0;
		PORTA = aux;
		PORTA |= (1 << PA5);
		for (i = 0; i < SIZE; i++) {
			for (j = 0; j < SIZE; j++) {
				if (board[i][j] == 1 || board[i][j] == 4 || board[i][j] == 5 || board[i][j] == 7) {
					if (i < 2)
						PORTB |= (1 << (i * SIZE + j));
					else
						PORTD |= (1 << ((i-2) * SIZE + j));
				}
			}
		}
		_delay_ms(10);
		PORTB = 0;
		PORTD = 0;
		PORTA = aux;
		PORTA |= (1 << PA6);
		for (i = 0; i < SIZE; i++) {
			for (j = 0; j < SIZE; j++) {
				if (board[i][j] == 2 || board[i][j] == 4 || board[i][j] == 6 || board[i][j] == 7) {
					if (i < 2)
						PORTB |= (1 << (i * SIZE + j));
					else
						PORTD |= (1 << ((i-2) * SIZE + j));
				}
			}
		}
		_delay_ms(10);
		PORTB = 0;
		PORTD = 0;
		PORTA = aux;
		PORTA |= (1 << PA7);
		for (i = 0; i < SIZE; i++) {
			for (j = 0; j < SIZE; j++) {
				if (board[i][j] == 3 || board[i][j] == 5 || board[i][j] == 6 || board[i][j] == 7) {
					if (i < 2)
						PORTB |= (1 << (i * SIZE + j));
					else
						PORTD |= (1 << ((i-2) * SIZE + j));
				}
			}
		}
		
		if ((PINA & (1 << PA1)) == 0) {
			if (max == 0) 
				success = moveLeft(board);
			prim = true;
			max = 25;
		}
		if ((PINA & (1 << PA2)) == 0) {
			if (max == 0) 
				success = moveDown(board);
			prim = true;
			max = 25;
		}
		if ((PINA & (1 << PA3)) == 0) {
			if (max == 0) 
				success = moveUp(board);
			prim = true;
			max = 25;
		}
		if ((PINA & (1 << PA4)) == 0) {
			if (max == 0) 
				success = moveRight(board);
			prim = true;
			max = 25;
		}
		
		if (max > 0) {
			max--;
			continue;
		}
		if (prim){
			f = gameEnded(board);
			if (success)
				addRandom(board);
			if (f > 0) {
				PORTB = 255;
				PORTD = 255;
				if (f == 1) {
					PORTA = aux;
					for (i = 0; i < 6; i++) {
						PORTA ^= (1 << PA5);
						_delay_ms(1000);
					}
					PORTA = aux;
				}
				else {
					PORTA = aux;
					for (i = 0; i < 6; i++) {
						PORTA ^= (1 << PA6);
						_delay_ms(1000);
					}
					PORTA = aux;
				}
				initBoard(board);
			}
			success = false;
		}
		
	}

}


